/**
 * Bull SAS / OW2 Orchestra
 * Copyright (C) 2010 Bull S.A.S, and individual contributors as indicated
 * by the @authors tag.
 *
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.designer.models.transformers.bpmn2designer;

import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import javax.xml.XMLConstants;
import javax.xml.bind.JAXBElement;
import javax.xml.namespace.QName;
import org.ow2.orchestra.designer.bpmn.model.FlowModel;
import org.ow2.orchestra.designer.bpmn.model.data.AbstractExpressionElementModel;
import org.ow2.orchestra.designer.bpmn.model.data.DataInputModel;
import org.ow2.orchestra.designer.bpmn.model.data.DataMappingModel;
import org.ow2.orchestra.designer.bpmn.model.data.DataObjectModel;
import org.ow2.orchestra.designer.bpmn.model.data.ExpressionDataInputModel;
import org.ow2.orchestra.designer.bpmn.model.data.ExpressionDataObjectModel;
import org.ow2.orchestra.designer.bpmn.model.data.type.DataTypeModel;
import org.ow2.orchestra.designer.bpmn.model.process.AbstractBPMNElementWithData;
import org.ow2.orchestra.designer.models.exceptions.TransformationException;
import org.ow2.orchestra.jaxb.bpmn.Assignment;
import org.ow2.orchestra.jaxb.bpmn.DataInput;
import org.ow2.orchestra.jaxb.bpmn.DataInputAssociation;
import org.ow2.orchestra.jaxb.bpmn.TDataObject;
import org.ow2.orchestra.jaxb.bpmn.TExpression;

/**
 * Transform {@link DataInputAssociation}s into {@link org.ow2.orchestra.designer.bpmn.model.data.DataMappingModel}s
 *
 * @author Guillaume Porcher
 */
public final class DataMapperTransformer {

  /**
   * Utility classes should not have public or default constructors.
   */
  private DataMapperTransformer() {

  }

  /**
   * @throws org.ow2.orchestra.designer.models.exceptions.TransformationException
   *
   */
  private static void bpmnToModel(
      final DataInputAssociation dataInputAssociation,
      final DataMappingModel dataMappingModel,
      final AbstractBPMNElementWithData elementWithData,
      final Bpmn2DesignerContext context,
      final int position) throws TransformationException {
    for (Assignment assignment : dataInputAssociation.getAssignments()) {
      // Parse to expr
      ExpressionDataInputModel expressionDataInputModel =
          getToExpressionModel(assignment.getTo(), dataMappingModel, elementWithData, position);
      // Parse from expr
      AbstractExpressionElementModel fromExpressionModel =
              getFromExpressionModel(assignment.getFrom(), dataInputAssociation, dataMappingModel, context, position);
      if (fromExpressionModel == null) {
        throw new TransformationException("Unable to generate expression.");
      }
      // Create link between from expr and to expr
      FlowModel flowModel = new FlowModel();
      flowModel.setSourceElement(fromExpressionModel);
      fromExpressionModel.getOutgoingFlows().add(flowModel);
      flowModel.setTargetElement(expressionDataInputModel);
      expressionDataInputModel.getIncomingFlows().add(flowModel);
      dataMappingModel.getElements().add(flowModel);
    }
  }

  private static ExpressionDataInputModel getToExpressionModel(final TExpression toExpression,
                                                               final DataMappingModel dataMappingModel,
                                                               final AbstractBPMNElementWithData elementWithData,
                                                               final int position) {
    // Parse to expr
    String toExpr = "";
    for (final Object content : toExpression.getContent()) {
      if (content instanceof String && !"".equals(content)) {
        toExpr = (String) content;
      }
    }
    ExpressionDataInputModel expressionDataInputModel = new ExpressionDataInputModel();
    // TODO layout
    expressionDataInputModel.setX(400);
    expressionDataInputModel.setY(50 * position);

    DataInputModel dataInputModel = null;
    String childNodeSelector = null;
    // Compile and use regular expression
    String dataInputPatternStr = "getDataInput\\(\"(.*)\"\\)/(.*)";
    Pattern pattern = Pattern.compile(dataInputPatternStr);
    Matcher matcher = pattern.matcher(toExpr);
    boolean matchFound = matcher.find();
    if (matchFound) {
      String dataInputName = matcher.group(1);
      String childNS = matcher.group(2);
      String[] tmp = childNS.split(":");
      if (tmp.length == 2) {
        String prefix = tmp[0];
        String namespace = toExpression.getOtherAttributes().get(new QName(XMLConstants.XMLNS_ATTRIBUTE_NS_URI, prefix));
        childNodeSelector = "{" + namespace + "}" + tmp[1];
      } else {
        childNodeSelector = childNS;
      }
      for (DataInputModel dim : elementWithData.getDataInputs()) {
        if (dim.getName().equals(dataInputName)) {
          dataInputModel = dim;
        }
      }
    } else {
      String dataInputPatternStr2 = "getDataInput\\(\"(.*)\"\\)";
      Pattern pattern2 = Pattern.compile(dataInputPatternStr2);
      Matcher matcher2 = pattern2.matcher(toExpr);
      boolean matchFound2 = matcher2.find();
      if (matchFound2) {
        String dataInputName = matcher2.group(1);
        for (DataInputModel dim : elementWithData.getDataInputs()) {
          if (dim.getName().equals(dataInputName)) {
            dataInputModel = dim;
          }
        }
      }
    }
    // set data input
    expressionDataInputModel.setDataInputModel(dataInputModel);
    expressionDataInputModel.setChildNodeSelector(childNodeSelector);
    dataMappingModel.getElements().add(expressionDataInputModel);
    return expressionDataInputModel;
  }

  private static AbstractExpressionElementModel getFromExpressionModel(
      final TExpression fromExpression,
      final DataInputAssociation dataInputAssociation,
      final DataMappingModel dataMappingModel,
      final Bpmn2DesignerContext context,
      final int position) throws TransformationException {
    String fromExpr = "";
    for (final Object content : fromExpression.getContent()) {
      if (content instanceof String && !"".equals(content)) {
        fromExpr = (String) content;
      }
    }
    DataObjectModel dataObjectModel = null;
    String childNodeSelector = null;
    // Compile and use regular expression for DataObject
    String dataObjectPatternStr = "getDataObject\\(\"(.*)\"\\)/(.*)";
    Pattern pattern = Pattern.compile(dataObjectPatternStr);
    Matcher matcher = pattern.matcher(fromExpr);
    boolean matchFound = matcher.find();
    if (matchFound) {
      String dataObjectName = matcher.group(1);
      String childNS = matcher.group(2);
      String[] tmp = childNS.split(":");
      if (tmp.length == 2) {
        String prefix = tmp[0];
        String namespace = fromExpression.getOtherAttributes().get(new QName(XMLConstants.XMLNS_ATTRIBUTE_NS_URI, prefix));
        childNodeSelector = "{" + namespace + "}" + tmp[1];
      } else {
        childNodeSelector = childNS;
      }
      // get DataObjectModel from Process
      for (JAXBElement<Object> jaxbSourceRef : dataInputAssociation.getSourceReves()) {
        if (jaxbSourceRef.getValue() instanceof TDataObject) {
          TDataObject dataObject = (TDataObject) jaxbSourceRef.getValue();
          if (dataObject.getName().equals(dataObjectName)) {
            ElementTransformer.bpmnToModel(dataObject, context);
            dataObjectModel = (DataObjectModel) context.getModelsMap().get(dataObject);
          }
        }
      }
      if (dataObjectModel == null) {
        throw new TransformationException("No dataObject '" + dataObjectName + "' found in expression " + fromExpr);
      }
    } else {
      String dataObjectPatternStr2 = "getDataObject\\(\"(.*)\"\\)";
      Pattern pattern2 = Pattern.compile(dataObjectPatternStr2);
      Matcher matcher2 = pattern2.matcher(fromExpr);
      boolean matchFound2 = matcher2.find();
      if (matchFound2) {
        String dataObjectName = matcher2.group(1);
        // get DataObjectModel from Process
        for (JAXBElement<Object> jaxbSourceRef : dataInputAssociation.getSourceReves()) {
          if (jaxbSourceRef.getValue() instanceof TDataObject) {
            TDataObject dataObject = (TDataObject) jaxbSourceRef.getValue();
            if (dataObject.getName().equals(dataObjectName)) {
              ElementTransformer.bpmnToModel(dataObject, context);
              dataObjectModel = (DataObjectModel) context.getModelsMap().get(dataObject);
            }
          }
        }
        if (dataObjectModel == null) {
          throw new TransformationException("No dataObject '" + dataObjectName + "' found in expression " + fromExpr);
        }
      }
    }
    if (dataObjectModel != null) {
      ExpressionDataObjectModel expressionDataObjectModel = new ExpressionDataObjectModel();
      // TODO layout
      expressionDataObjectModel.setX(50);
      expressionDataObjectModel.setY(50 * position);

      expressionDataObjectModel.setDataObjectModel(dataObjectModel);
      expressionDataObjectModel.setChildNodeSelector(childNodeSelector);
      dataMappingModel.getElements().add(expressionDataObjectModel);
      return expressionDataObjectModel;
    }
    // If getDataObject not found, look for another construction (operation ?)
    return null;
  }


  public static void bpmnToModel(
      final List<DataInputAssociation> dataInputAssociations,
      final AbstractBPMNElementWithData elementWithData,
      final Bpmn2DesignerContext context) throws TransformationException {
    DataMappingModel dataMappingModel = new DataMappingModel();
    int position = 0;
    for (DataInputAssociation dataInputAssociation : dataInputAssociations) {
      position++;
      bpmnToModel(dataInputAssociation, dataMappingModel, elementWithData, context, position);
    }
    elementWithData.setDataMappingModel(dataMappingModel);
  }

  public static void getDataInputModels(final AbstractBPMNElementWithData elementWithData,
                                        final List<DataInput> dataInputs,
                                        final Bpmn2DesignerContext context) {
    for (DataInput dataInput : dataInputs) {
      DataInputModel dataInputModel = new DataInputModel();
      dataInputModel.setId(dataInput.getId());
      dataInputModel.setName(dataInput.getName());
      String itemDefinitionId = dataInput.getItemSubjectRef().getLocalPart(); // TODO handle multiple namespaces
      DataTypeModel dataTypeModel = ElementTransformer.getDataTypeModel(context, itemDefinitionId);

      dataInputModel.setType(dataTypeModel);
      dataInputModel.setElementModel(elementWithData);
      elementWithData.getDataInputs().add(dataInputModel);
    }
  }

}
